use crate::ast::{BinaryType, Expr, Node, UnaryType};
use crate::tokens::{Position, PositionKnowing, Token, TokenType};

#[derive(Debug)]
pub struct ParserError {
    pos: Position,
    message: String,
}

pub struct Parser<'a, Tokens>
where
    Tokens: PositionKnowing + Iterator<Item = Token>,
{
    text: &'a str,
    tokens: Tokens,
    current_token: Option<Token>,
}

impl<'a, Tokens> Parser<'a, Tokens>
where
    Tokens: PositionKnowing + Iterator<Item = Token>,
{
    pub fn new(text: &'a str, mut lexer: Tokens) -> Self {
        Self {
            text,
            current_token: lexer.next(),
            tokens: lexer,
        }
    }

    pub fn parse_expr(&mut self) -> Result<Node<Expr>, ParserError> {
        self.parse_prec_or()
    }

    fn parse_prec_or(&mut self) -> Result<Node<Expr>, ParserError> {
        let mut left = self.parse_prec_and()?;
        while !self.done() {
            if self.current_is(TokenType::Or) {
                self.step();
                let right = self.parse_prec_and()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Or,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        Ok(left)
    }

    fn parse_prec_and(&mut self) -> Result<Node<Expr>, ParserError> {
        let mut left = self.parse_prec_equal_inequal()?;
        while !self.done() {
            if self.current_is(TokenType::And) {
                self.step();
                let right = self.parse_prec_equal_inequal()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::And,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        Ok(left)
    }

    fn parse_prec_equal_inequal(&mut self) -> Result<Node<Expr>, ParserError> {
        let mut left = self.parse_prec_lt_lte_gt_gte_in()?;
        while !self.done() {
            if self.current_is(TokenType::EqualEqual) {
                self.step();
                let right = self.parse_prec_lt_lte_gt_gte_in()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Equal,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::ExclamationEqual) {
                self.step();
                let right = self.parse_prec_lt_lte_gt_gte_in()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Inequal,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        Ok(left)
    }

    fn parse_prec_lt_lte_gt_gte_in(&mut self) -> Result<Node<Expr>, ParserError> {
        let mut left = self.parse_prec_add_subtract()?;
        while !self.done() {
            if self.current_is(TokenType::LessThan) {
                self.step();
                let right = self.parse_prec_add_subtract()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::LT,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::GreaterThan) {
                self.step();
                let right = self.parse_prec_add_subtract()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::GT,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::LessThanEqual) {
                self.step();
                let right = self.parse_prec_add_subtract()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::LTE,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::GreaterThanEqual) {
                self.step();
                let right = self.parse_prec_add_subtract()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::GTE,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::In) {
                self.step();
                let right = self.parse_prec_add_subtract()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::In,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        Ok(left)
    }

    fn parse_prec_add_subtract(&mut self) -> Result<Node<Expr>, ParserError> {
        let mut left = self.parse_prec_multiply_divide_modulo()?;
        while !self.done() {
            if self.current_is(TokenType::Plus) {
                self.step();
                let right = self.parse_prec_multiply_divide_modulo()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Add,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::Minus) {
                self.step();
                let right = self.parse_prec_multiply_divide_modulo()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Subtract,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        Ok(left)
    }

    fn parse_prec_multiply_divide_modulo(&mut self) -> Result<Node<Expr>, ParserError> {
        let mut left = self.parse_prec_unary()?;
        while !self.done() {
            if self.current_is(TokenType::Asterisk) {
                self.step();
                let right = self.parse_prec_unary()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Multiply,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::Slash) {
                self.step();
                let right = self.parse_prec_unary()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Divide,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else if self.current_is(TokenType::Percent) {
                self.step();
                let right = self.parse_prec_unary()?;
                left = self.node(Expr::Binary {
                    binary_type: BinaryType::Modulo,
                    left: Box::new(left),
                    right: Box::new(right),
                });
            } else {
                break;
            }
        }
        Ok(left)
    }

    fn parse_prec_unary(&mut self) -> Result<Node<Expr>, ParserError> {
        if !self.done() && self.current_is(TokenType::Not) {
            self.step();
            let subject = Box::new(self.parse_prec_unary()?);
            self.ok_node(Expr::Unary {
                unary_type: UnaryType::Not,
                subject,
            })
        } else if !self.done() && self.current_is(TokenType::Minus) {
            self.step();
            let subject = Box::new(self.parse_prec_unary()?);
            self.ok_node(Expr::Unary {
                unary_type: UnaryType::Negate,
                subject,
            })
        } else if !self.done() && self.current_is(TokenType::Ampersand) {
            self.step();
            if !self.done() && self.current_is(TokenType::Mut) {
                self.step();
                let subject = Box::new(self.parse_prec_unary()?);
                self.ok_node(Expr::Unary {
                    unary_type: UnaryType::ReferenceMut,
                    subject,
                })
            } else {
                let subject = Box::new(self.parse_prec_unary()?);
                self.ok_node(Expr::Unary {
                    unary_type: UnaryType::Reference,
                    subject,
                })
            }
        } else if !self.done() && self.current_is(TokenType::Asterisk) {
            self.step();
            let subject = Box::new(self.parse_prec_unary()?);
            self.ok_node(Expr::Unary {
                unary_type: UnaryType::Dereference,
                subject,
            })
        } else {
            self.parse_prec_exponentiate()
        }
    }

    fn parse_prec_exponentiate(&mut self) -> Result<Node<Expr>, ParserError> {
        let left = self.parse_prec_member_index_call()?;
        if !self.done() && self.current_is(TokenType::AsteriskEqual) {
            let right = self.parse_prec_exponentiate()?;
            self.step_and_ok_node(Expr::Binary {
                binary_type: BinaryType::Exponentiate,
                left: Box::new(left),
                right: Box::new(right),
            })
        } else {
            Ok(left)
        }
    }

    fn parse_prec_member_index_call(&mut self) -> Result<Node<Expr>, ParserError> {
        let mut subject = self.parse_operand()?;
        while !self.done() {
            if self.current_is(TokenType::Dot) {
                self.step();
                if self.done() || self.current_is(TokenType::Id) {
                    return self.error("expected identifier");
                }
                let value = self.token_string(self.current());
                self.step();
                subject = self.node(Expr::Member {
                    subject: Box::new(subject),
                    value,
                });
            } else if self.current_is(TokenType::LBracket) {
                self.step();
                let value = self.parse_expr()?;
                if self.done() || !self.current_is(TokenType::RBracket) {
                    return self.error("expected ']'");
                }
                subject = self.node(Expr::Index {
                    subject: Box::new(subject),
                    value: Box::new(value),
                });
            } else if self.current_is(TokenType::LParen) {
                self.step();
                let mut arguments = Vec::<Node<Expr>>::new();
                if !self.done() && !self.current_is(TokenType::RParen) {
                    arguments.push(self.parse_expr()?);
                    while !self.done() && self.current_is(TokenType::Comma) {
                        self.step();
                        if self.done() || self.current_is(TokenType::RParen) {
                            self.step();
                        }
                        arguments.push(self.parse_expr()?);
                    }
                }
                if self.done() || !self.current_is(TokenType::RParen) {
                    return self.error("expected ')'");
                }
                self.step();
                subject = self.node(Expr::Call {
                    subject: Box::new(subject),
                    arguments,
                })
            } else {
                break;
            }
        }
        Ok(subject)
    }

    fn parse_operand(&mut self) -> Result<Node<Expr>, ParserError> {
        if self.done() {
            return self.error("expected value, got eof");
        }
        match self.current().token_type {
            TokenType::Id => self.step_and_ok_node(Expr::Id(self.token_string(self.current()))),
            TokenType::Int => {
                let mut value_string = self.token_string(self.current());
                self.step();
                if !self.done() && self.current_is(TokenType::Decimal) {
                    value_string.push_str(&self.token_string(self.current()));
                    self.step_and_ok_node(Expr::Float(
                        value_string.parse::<f64>().expect("valid f64"),
                    ))
                } else {
                    self.ok_node(Expr::Int((value_string).parse::<i64>().expect("valid i64")))
                }
            }
            TokenType::Decimal => self.step_and_ok_node(Expr::Float(
                self.token_string(self.current())
                    .parse::<f64>()
                    .expect("valid f64"),
            )),
            TokenType::False => self.step_and_ok_node(Expr::Bool(false)),
            TokenType::True => self.step_and_ok_node(Expr::Bool(true)),
            TokenType::LParen => self.parse_unit_group_or_tuple(),
            TokenType::LBrace => self.parse_object(),
            TokenType::LBracket => self.parse_array(),
            TokenType::Fn => self.parse_function(),
            _ => self.error("expected value"),
        }
    }

    fn parse_unit_group_or_tuple(&mut self) -> Result<Node<Expr>, ParserError> {
        self.step();
        if !self.done() && !self.current_is(TokenType::LParen) {
            todo!()
        } else {
            self.step_and_ok_node(Expr::Unit)
        }
    }

    fn parse_object(&mut self) -> Result<Node<Expr>, ParserError> {
        todo!()
    }

    fn parse_array(&mut self) -> Result<Node<Expr>, ParserError> {
        todo!()
    }

    fn parse_function(&mut self) -> Result<Node<Expr>, ParserError> {
        todo!()
    }

    fn token_string(&self, token: &Token) -> String {
        self.text[token.pos.index..token.pos.index + token.length].to_string()
    }

    fn step_and_ok_node<T>(&mut self, value: T) -> Result<Node<T>, ParserError> {
        self.step();
        self.ok_node(value)
    }

    fn ok_node<T>(&self, value: T) -> Result<Node<T>, ParserError> {
        Ok(Node {
            value,
            pos: self.tokens.pos(),
        })
    }

    fn step_and_node<T>(&mut self, value: T) -> Node<T> {
        self.step();
        self.node(value)
    }

    fn node<T>(&self, value: T) -> Node<T> {
        Node {
            value,
            pos: self.tokens.pos(),
        }
    }

    fn step_and<T>(&mut self, value: T) -> T {
        self.step();
        value
    }

    fn error(&self, message: &str) -> Result<Node<Expr>, ParserError> {
        Err(ParserError {
            pos: self.tokens.pos(),
            message: message.to_string(),
        })
    }

    fn done(&self) -> bool {
        self.current_token.is_none()
    }

    fn current_is(&self, token_type: TokenType) -> bool {
        self.current().token_type == token_type
    }

    fn current(&self) -> &Token {
        self.current_token.as_ref().expect("done() checked")
    }

    fn step(&mut self) {
        self.current_token = self.tokens.next();
    }
}
